from dataclasses import dataclass
from typing import Optional

import numpy as np
from colour import Color
from PIL import Image


@dataclass
class Settings:
    """Tools settings"""

    saving: Optional[int] = None
    """Defines what part of the pixels will be saved (not erased).
    From 0 (erase everything) to 100 (save everything)"""

    thickness: int = 3
    """Thickness of the line, used in algorithm to set blur radius"""

    color: Color = Color("black")
    """Color of the contour"""

    crop: bool = True
    """Crop empty pixels after converting"""


@dataclass
class ImageInfo:
    """Describes image with additional information"""

    original: Image
    """Original images"""

    settings: Settings
    """Settings that were used to process image"""

    bool_array: np.ndarray
    """2D array that represents used and unused pixels"""



from pathlib import Path
from typing import NamedTuple

from colour import Color


class Settings(NamedTuple):
    """Tools settings"""

    in_file: Path
    """Input file"""

    out_file: Path
    """Output file"""

    lower: int
    """Lower threshold for Candy edge detection"""

    upper: float
    """Upper threshold for Candy edge detection"""

    color: Color
    """Final color of contour"""

    crop: bool
    """Crop transparent pixels after converting"""



import os
from pathlib import Path

import click
from colour import Color
from loguru import logger
from PIL import Image

from monochromist.math_version.clean import erase
from monochromist.math_version.postprocess import color_and_crop

from .classes import ImageInfo, Settings


@click.command()
@click.option(
    "-i", "--input", type=click.Path(exists=True, dir_okay=False), required=True, help="Input file"
)
@click.option(
    "-o",
    "--output",
    type=click.Path(writable=True, dir_okay=False),
    required=True,
    help="Output file",
)
@click.option(
    "-s",
    "--saving",
    type=click.IntRange(min=0, max=100),
    help="How much pixels to save, from 0 (erase all) to 100 (save all). "
    "If not defined then empirical algorithm will be used",
)
@click.option(
    "-t",
    "--thickness",
    type=int,
    help="Thickness of the line",
    default=3,
    show_default=True,
)
@click.option(
    "-c",
    "--color",
    type=str,
    help="Name of the color or it's HEX code (ex. #CC573E)",
    default="black",
    show_default=True,
)
@click.option(
    "-r",
    "--crop",
    type=bool,
    help="Drop transparent pixels after image processing",
    default=True,
    show_default=True,
)
def process(
    input: Path, output: Path, saving: int, thickness: int, color: Color, crop: bool
) -> None:
    """Take contour from selected file"""

    parsed_color = Color(color)

    settings = Settings(
        thickness=thickness,
        saving=saving,
        color=parsed_color,
        crop=crop,
    )

    process_file(input, output, settings)


def process_file(input: Path, output: Path, settings: Settings) -> ImageInfo:
    """Take contour from selected file"""
    initial_image = Image.open(input)
    new_image, image_info = process_image(initial_image, settings)

    if os.path.exists(output):
        os.remove(output)
    new_image.save(output)

    logger.info(f"Convert <{input}> to <{output}>")
    logger.info(f"Used settings: {image_info.settings}")
    if settings.saving is None:
        logger.info("To better result try to vary saving parameter"
                    "by adding `-s <from 0 to 100>` to the end of the command")
    return image_info


def process_image(initial_image: Image, settings: Settings) -> tuple[Image, ImageInfo]:
    image_info = erase(initial_image, settings)
    new_image = color_and_crop(image_info)
    return new_image, image_info


if __name__ == "__main__":
    process()



from copy import copy

import numpy as np
from PIL import Image, ImageFilter

from .classes import ImageInfo, Settings


def erase(img: Image, settings: Settings) -> ImageInfo:
    """Erases pixels and returns numpy array the same shape as image,
    filled with True (save pixel) and False (erase pixel), and settings that were used.
    Parameters will be used if they are defined, otherwise empirical approach will be used."""
    # Convert to grayscale
    converted = img.convert("L")

    # Blur filter is used to average background
    blured = converted.filter(ImageFilter.MedianFilter(settings.thickness))
    values = np.asarray(blured)

    updated_settings = copy(settings)
    if not settings.saving is None:
        threshold = user_percentile(values, settings)
    else:
        saving, threshold = empirical_percentile(values)
        updated_settings.saving = saving

    erased = values < threshold

    return ImageInfo(img, updated_settings, erased)


def user_percentile(arr: np.ndarray, settings: Settings) -> int:
    saving_safe_value = max(min(settings.saving, 100), 0)
    return np.asscalar(np.percentile(arr.flatten(), saving_safe_value, axis=0))


def empirical_percentile(arr: np.ndarray) -> tuple[int, int]:
    percentiles = [np.percentile(arr.flatten(), i, axis=0) for i in range(101)]
    # y = K * x + b
    K = 1.5
    b_values = np.array([y - K * x for (x, y) in enumerate(percentiles)])
    optimal_index = np.asscalar(np.argmax(b_values))
    optimal_percentile = percentiles[optimal_index]

    return optimal_index, optimal_percentile



import numpy as np
from colour import Color
from PIL import Image

from monochromist.math_version.classes import ImageInfo


def color_and_crop(image_info: ImageInfo) -> Image:
    """Color non-transparent pixels with selected color, then crop transparent pixels if needed"""

    color_as_tuple = color_to_tuple(image_info.settings.color)
    transparent = [0, 0, 0, 0]

    def color_row(row: np.array) -> np.array:
        return np.array([color_as_tuple if x else transparent for x in row])

    colored = np.apply_along_axis(color_row, 1, image_info.bool_array)
    new_image = Image.fromarray(np.uint8(colored), "RGBA")

    if image_info.settings.crop and image_info.bool_array.any():
        borders = find_borders(image_info.bool_array)
        new_image = new_image.crop(borders)

    return new_image


def color_to_tuple(color: Color) -> list[int]:
    """Convert color to RGBA tuple"""
    r, g, b = [int(255 * x) for x in color.rgb]
    alpha_channel = 255
    return [r, g, b, alpha_channel]


def find_borders(arr: np.array) -> tuple[int, int, int, int]:
    """Find transparent borders"""

    notna_columns = arr.any(axis=0)
    notna_rows = arr.any(axis=1)

    left = np.flatnonzero(notna_columns)[0]
    right = np.flatnonzero(notna_columns)[-1]

    upper = np.flatnonzero(notna_rows)[0]
    lower = np.flatnonzero(notna_rows)[-1]

    return left, upper, right, lower
